MODULE TEST_lib_divrot

  USE FORTUTF
  use, intrinsic :: iso_fortran_env, only: wp => real64
  USE mo_iconlib_kind, ONLY: dp
  USE mo_lib_divrot

  IMPLICIT NONE

  PRIVATE

  PUBLIC :: TEST_recon_lsq_cell_l_lib

  CONTAINS

  SUBROUTINE TEST_recon_lsq_cell_l_lib

    ! Parameters for the test case
    integer, parameter :: nblk = 1
    integer, parameter :: nproma = 3   ! number of cells in our simplified case
    integer, parameter :: nvert = 1    ! vertical levels
    integer, parameter :: lsq_dim_unk = 2
    integer, parameter :: lsq_dim_c = 3

    ! Local indices for the stencil indices
    integer, parameter :: i_startblk = 1, i_endblk = nblk
    integer, parameter :: i_startidx_in = 1, i_endidx_in = nproma
    integer, parameter :: slev = 1, elev = nvert

    ! Declarations for input arrays
    real(wp), allocatable :: p_cc(:,:,:)
    integer, allocatable, target :: cell_neighbor_idx(:,:,:)
    integer, allocatable, target :: cell_neighbor_blk(:,:,:)
    real(wp), allocatable :: lsq_qtmat_c(:,:,:,:)
    real(wp), allocatable :: lsq_rmat_rdiag_c(:,:,:)
    real(wp), allocatable :: lsq_rmat_utri_c(:,:,:)
    real(wp), allocatable :: lsq_moments(:,:,:)
    real(wp), allocatable :: p_coeff(:,:,:,:)  ! output

    ! Loop variables
    integer :: i, j, k

    ! Tolerance for comparisons
    real(wp), parameter :: tol = 1.0e-6_wp

    ! Expected values for one cell (cell 1 in block 1 and level 1)
    real(wp) :: expected_const, expected_zonal, expected_meridional

    ! Allocate arrays
    allocate(p_cc(nproma, nvert, nblk))
    allocate(cell_neighbor_idx(nproma, nblk, 3))
    allocate(cell_neighbor_blk(nproma, nblk, 3))
    allocate(lsq_qtmat_c(nproma, lsq_dim_unk, lsq_dim_c, nblk))
    allocate(lsq_rmat_rdiag_c(nproma, lsq_dim_unk, nblk))
    allocate(lsq_rmat_utri_c(nproma, (lsq_dim_unk*lsq_dim_unk - lsq_dim_unk)/2, nblk))
    allocate(lsq_moments(nproma, nblk, lsq_dim_unk))
    allocate(p_coeff(3, nproma, nvert, nblk))

    ! Initialize p_cc: let each cell have a constant value equal to its cell index.
    do i = 1, nproma
      p_cc(i,1,1) = real(i, wp)
    end do

    ! Initialize cell_neighbor_idx and cell_neighbor_blk.
    ! For simplicity, let all three neighbor indices for each cell be:
    ! neighbor 1: next cell (cyclic), neighbor 2: same cell, neighbor 3: same cell.
    do i = 1, nproma
      cell_neighbor_idx(i,1,1) = mod(i, nproma) + 1
      cell_neighbor_idx(i,1,2) = i
      cell_neighbor_idx(i,1,3) = i
      cell_neighbor_blk(i,1,:) = 1   ! All from block 1
    end do

    ! For simplicity, let for any cell:
    !   lsq_qtmat_c(cell,1,k,1) = 1.0  and lsq_qtmat_c(cell,2,k,1) = 0.5   for k=1,2,3.
    do i = 1, nproma
      do k = 1, lsq_dim_c
        lsq_qtmat_c(i,1,k,1) = 1.0_wp
        lsq_qtmat_c(i,2,k,1) = 0.5_wp
      end do
    end do

    ! Initialize lsq_rmat_rdiag_c.
    do i = 1, nproma
      lsq_rmat_rdiag_c(i,1,1) = 2.0_wp
      lsq_rmat_rdiag_c(i,2,1) = 2.0_wp
    end do

    ! Initialize lsq_rmat_utri_c.
    ! For each cell, set the single value (for 2 unknowns) to 0.1.
    do i = 1, nproma
      lsq_rmat_utri_c(i,1,1) = 0.1_wp
    end do

    ! Initialize lsq_moments.
    ! For each cell, let the first moment be 0.2 and the second moment be 0.3.
    do i = 1, nproma
      lsq_moments(i,1,1) = 0.2_wp
      lsq_moments(i,1,2) = 0.3_wp
    end do

    ! Initialize p_coeff to zero
    p_coeff = 0.0_wp

    ! Call the reconstruction subroutine.
    ! Note: In this test we set l_consv = .true. and do not use OpenACC options.
    call recon_lsq_cell_l_lib(p_cc, cell_neighbor_idx, cell_neighbor_blk,   &
    & lsq_qtmat_c, lsq_rmat_rdiag_c, lsq_rmat_utri_c, lsq_moments, p_coeff, &
    & i_startblk, i_endblk, i_startidx_in, i_endidx_in, &
    & slev, elev, nproma, .true.)

    ! Expected values for cell 1:
    expected_meridional = 1.0_wp
    expected_zonal      = 1.8_wp
    expected_const      = 0.34_wp

    if (abs(p_coeff(1,1,1,1) - expected_const) > tol) then
      print *, "Test failed: Constant coefficient (cell 1) =", p_coeff(1,1,1,1), &
      " Expected =", expected_const
      stop 1
    end if

    if (abs(p_coeff(2,1,1,1) - expected_zonal) > tol) then
      print *, "Test failed: Zonal gradient coefficient (cell 1) =", p_coeff(2,1,1,1), &
      " Expected =", expected_zonal
      stop 1
    end if

    if (abs(p_coeff(3,1,1,1) - expected_meridional) > tol) then
      print *, "Test failed: Meridional gradient coefficient (cell 1) =", p_coeff(3,1,1,1), &
      " Expected =", expected_meridional
      stop 1
    end if

    print *, "All tests passed successfully."

  END SUBROUTINE TEST_recon_lsq_cell_l_lib
END MODULE TEST_lib_divrot
